/*
 * Copyright (c) 2000, 2001 Vladimir Popov <jumbo@narod.ru>.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
/*
 * $Id: aztech.c,v 1.38 2001/05/15 09:54:52 pva Exp $
 * lowlevel io functions for fmio utility -- aztech driver
 */

#include <unistd.h>

#ifdef linux
#include <sys/io.h>
#elif defined __FreeBSD__
#include <fcntl.h>
#include <machine/cpufunc.h>
#else
#include <machine/pio.h>
#include <machine/sysarch.h>
#endif

#include "io.h"

static void send_zero(void);
static void send_one(void);

static int stereo = 1; /* Use stereo by default */
static int vol = 0;
static unsigned long radioport = 0;
#ifdef __FreeBSD__
static int iofd;
#endif /* __FreeBSD__ */

/******************************************************************/

int
Grab_Port_aztech(unsigned long port) {
#if defined (__OpenBSD__) || defined (__NetBSD__)
	unsigned long iomap[32];
	int offset;
	unsigned long mask;
	struct i386_set_ioperm_args ioperms;
#endif /* __OpenBSD__ || __NetBSD__ */

	radioport = port;

#ifdef linux /* Platform dependent */
	if ( ioperm(radioport, 1, 1) < 0 )
#elif defined __FreeBSD__
	if ( (iofd = open("/dev/io", O_RDONLY)) < 0 )
#else
	offset = radioport/32;
	mask = 1 << radioport%32;
	memset(iomap, 0xFFFF, sizeof(iomap));
	iomap[offset] ^= mask;
	ioperms.iomap = iomap;
	if ( sysarch(I386_SET_IOPERM, (char *)&ioperms) < 0 )
#endif /* Platform dependent */
		return 1;

	return 0;
}

int
Release_Port_aztech(void) {
#if defined (__OpenBSD__) || defined (__NetBSD__)
	unsigned long iomap[32];
	struct i386_set_ioperm_args ioperms;
#endif /* __OpenBSD__ || __NetBSD__ */

#ifdef linux /* Platform dependent */
	if ( ioperm(radioport, 1, 0) < 0 )
#elif defined __FreeBSD__
	if ( close(iofd) < 0 )
#else
	memset(iomap, 0xFFFF, sizeof(iomap));
	ioperms.iomap = iomap;
	if ( sysarch(I386_SET_IOPERM, (char *)&ioperms) < 0 )
#endif /* Platform dependent */
		return 1;

	return 0;
}

void
Change_Frequency_aztech(unsigned long frequency) {
	int i;
	int rf = 50; /* 50 kHz */

	frequency += 1070;	/* Add IF 10.7 MHz */
	frequency *= 10; /* Keep this, or the rf = 25 kHz will fail */
	frequency /= rf; /* Static divider for bandwidth */

	/* 14 frequency bits */
	for (i = 0; i < 14; i++)
		if ( frequency & (1<<i) )
			send_one();
		else
			send_zero();

	/* LSI test bits */
	send_zero();
	send_zero();

	/* Band */
	send_one();
	if ( stereo ) /* Send stereo bit */
		send_one();
	else
		send_zero();
	send_zero();

	/* Time base */
	send_zero();

	/* Reference frequency */
	switch ( rf ) {
		case 25: /* 25 kHz */
			send_zero();
			send_one();
			send_zero();
			break;
		case 100: /* 100 kHz */
			send_zero();
			send_zero();
			send_zero();
			break;
		case 50: /* 50 kHz */
		default:
			send_zero();
			send_zero();
			send_one();
			break;
	}

	/* Do FM */
	send_one();

	OUTB(radioport, 0x80+0x40+vol); /* Hey, we're done */

	return;
}

void
mute_aztech(void) {
	vol = 0;
	OUTB(radioport, 0);
}

/*
 * Volume is adjusted by manipulating bits 00000x0x,
 * that is, the only allowed values are 101(5), 100(4), 001(1), 000(0)
 */
void
volumeup_aztech(void) {
	if ( vol < 5 ) {
		switch ( vol ) {
			case 0: vol = 1; break;
			case 1: vol = 4; break;
			case 4: vol = 5; break;
		}

		OUTB(radioport, vol);
	}
}

void
volumedn_aztech(void) {
	if ( vol > 0 ) {
		switch ( vol ) {
			case 5: vol = 4; break;
			case 4: vol = 1; break;
			case 1: vol = 0; break;
		}

		OUTB(radioport, vol);
	}
}

int
state_aztech(void) {
	return inb(radioport) & 3;
}

void
mono_aztech(void) {
	stereo = 0;
}

static void
send_zero(void) {
	OUTB(radioport, 0x02+vol);
	OUTB(radioport, 0x40+0x02+vol);
}

static void
send_one(void) {
	OUTB(radioport, 0x80+0x02+vol);
	OUTB(radioport, 0x80+0x40+0x02+vol);
}
