/*
 * Copyright (c) 2000, 2001 Vladimir Popov <jumbo@narod.ru>.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
/*
 * $Id: fmio.c,v 1.62 2001/07/15 12:42:10 pva Exp $
 *
 * core file for fmio utility
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <err.h>

#include "config.h"
#include "radio.h"

/* Actions */
/* major */
#define NONE	0x00
#define SCAN	0x10
#define DETE	0x20
#define SRCH	0x40
/* minor */
#define MINOR	0x0F
#define STAT	0x01
#define TUNE	0x02
#define VOLU	0x04
#define MONO	0x08

void usage(void);
void invalid_driver_error(char *);
void die(int);

char *pn = NULL;
const char signal_found[] = "%s: %.2f MHz\n";

static int mixer = 0;

int
main(int argc, char *argv[]) {
	int optchar;
	int action = NONE;
	int volume = 0;
	char *drv = NULL;
	char *master_volume = NULL;
	char *other_volume = NULL;
	unsigned freq = 0;
	long search = 0;
	/* Scan mode doesn't actually need default values */
	unsigned lower = 0;
	unsigned higher = 0;
	unsigned long cycle = 1;

	pn = strrchr(argv[0], '/');
	if ( !pn )
		pn = argv[0];
	else
		pn++;

	/* 
	 * Call radio_init() before usage(),
	 * or default driver will be: NULL, 0x0
	 */
	drv = getenv("FMTUNER");
	if ( drv && *drv ) {
		if ( radio_init(drv) == ERADIO_INVL ) {
			invalid_driver_error(drv);
			radio_init(DEF_DRV);
		}
	} else {
		radio_init(DEF_DRV);
	}

	if ( argc < 2 ) {
		usage();
		exit(1);
	}

	/* Argh... options */
	while ( (optchar = getopt(argc, argv, "Dd:Sl:h:c:f:v:X:x:W:sm")) != -1 ) {
		switch ( optchar ) {
			case 'W':
				action = SRCH;
				search = atof(optarg)*100;
				break;
			case 'S': /* Scan mode */
				action = SCAN;
				break;
			case 'D': /* Detection mode */
				action = DETE;
				/* Don't clobber already set values */
				if ( higher == 0 ) higher = DEF_HIGHER_FREQ;
				if ( lower == 0 ) lower = DEF_LOWER_FREQ;
				break;
			case 'l': /* lower frequency of scan range */
				lower = atof(optarg)*100;
				break;
			case 'h': /* higher frequency of scan range */
				higher = atof(optarg)*100;
				break;
			case 'c': /* number of probes for each scanned frequency */
				if ( (cycle = (unsigned long)strtol(optarg, (char **)NULL, 10)) == 0 )
					cycle = 1;
				break;
			case 'v': /* set volume */
				action |= VOLU;
				volume = (int)strtol(optarg, (char **)NULL, 10);
				break;
			case 'X': /* set outputs.master */
				master_volume = optarg;
				break;
			case 'x':
				other_volume = optarg;
				break;
			case 'd': /* use this driver */
				if ( radio_init(optarg) == ERADIO_INVL )
					invalid_driver_error(optarg);
				break;
			case 'f': /* tune to the frequency */
				action |= TUNE;
				if ( (freq = atof(optarg)*100) == 0 )
					freq = DEF_FREQ;
				break;
			case 's': /* get signal quality */
				action |= STAT;
				break;
			case 'm': /* set card output to mono */
				action |= MONO;
				break;
			default:
				usage();
				exit(0);
		}
	}

	/* Minor actions have more priority */
	if ( action & MINOR ) action &= MINOR;

	/* Drop privs for drivers that don't need root */
	if ( (action & ~MINOR) != DETE )
		if ( !radio_info_root() )
			setuid(getuid());

	/* Enabling communication with the radio port */
	if ( radio_get_port() != 0 ) {
		warn("Port access error");
		exit(1);
	}

	/*
	 * During time consuming actions fmio can get
	 * some signal. Define action for such emergency.
	 * Though it's more suitable to define these under SCAN or DETE,
	 * it won't hurt if they'll be here.
	 */
	signal(SIGINT , die);
	signal(SIGHUP , die);
	signal(SIGTERM, die);

	switch ( action & ~MINOR ) {
		case NONE:
			if ( radio_mixer_init() != 0 )
				mixer = 0;
			else
				mixer = 1;

			if ( action & MONO ) radio_set_mono();
			if ( action & VOLU ) {
				radio_mute();
				while ( volume-- )
					radio_volume_up();
			}
			if ( action & TUNE ) { /* tune to the frequency freq */
				if ( action & VOLU ) {
					radio_set_frequency(freq);
				} else {
					radio_mute();
					radio_set_frequency(freq);
					radio_volume_up();
				}
			}

			if ( action & STAT ) { /* probe for signal sterngth */
				printf("%s: ", pn);
				switch( radio_get_state() ) {
					case 0:
						puts("stereo");
						break;
					case 1:
					case 2:
						puts("mono");
						break;
					default:
						puts("none");
						break;
				}
			}

			if ( mixer )
				if ( other_volume )
					radio_mixer_update_volume(other_volume);

			/*
			 * Now outputs.master=0,0
			 * Update saved values
			 */
			if ( mixer )
				if ( master_volume )
					radio_mixer_set_volume(master_volume);

			if ( mixer )
				radio_mixer_cleanup();
			break;
		case SCAN: /* Scan specified range */
			radio_scan(lower, higher, cycle, 1, radio_info_drv());
			break;
		case DETE: /* detection of all radio cards */
			radio_detect(lower, higher, cycle);
			break;
		case SRCH:
			if ( radio_mixer_init() != 0)
				mixer = 0;
			else
				mixer = 1;

			if ( search < 0 )
				freq = radio_search_down(-1*search);
			else
				freq = radio_search_up(search);
			if ( freq )
				printf(signal_found, pn, (float)freq/100);

			do {
				radio_volume_up();
			} while ( volume-- );

			if ( mixer )
				radio_mixer_cleanup();
			break;
	}

	/* Disabling communication with the radio port */
	if ( radio_free_port() != 0 )
		warn("Port release error");

	return 0;
}

/*
 * Dump the usage string to screen
 */
void
usage(void) {
const char usage_string[] =	"%s version %s\n"
				"Default driver: %s, port 0x%x\n"
				"Usage:\n"
				"  %s [-d driver] [-f frequency] [-v volume] [-x volume] [-X volume] [-s] [-m]\n" 
				"  %s [-d driver] -S [-l start -h end -c cycles]\n"
				"  %s [-d driver] -W frequency\n"
				"  %s -D [-l start -h end -c cycles]\n";

	printf(usage_string, pn, VERSION, radio_info_name(), radio_info_port(),
			pn, pn, pn, pn);
	return;
}

/*
 * Complain about invalid driver
 */
void
invalid_driver_error(char *invl_drv) {
	const char invl_drv_string[] = "%s: Invalid driver %s, using default %s\n";
	fprintf(stderr, pn, invl_drv_string, invl_drv, DEF_DRV);
}

void
die(int sig) {
	if ( mixer )
		radio_mixer_cleanup();

	if ( radio_free_port() != 0 )
		warn("Port release error");

	exit(1);
}
