/*
 * Copyright (c) 2001 Vladimir Popov <jumbo@narod.ru>.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
/*
 * $Id: maxiradio.c,v 1.17 2001/05/15 10:23:33 pva Exp $
 * TEA5757: http://www.semiconductors.com/acrobat/datasheets/TEA5757_5759_3.pdf
 * lowlevel io functions for fmio utility -- maxiradio driver
 * Based on a driver written by Jeremie Andrei <jimee@free.fr>
 */

#include <unistd.h>

#ifdef linux
#include <sys/io.h>
#elif defined __FreeBSD__
#include <fcntl.h>
#include <machine/cpufunc.h>
#else
#include <machine/pio.h>
#include <machine/sysarch.h>
#endif

#include "io.h"

#ifndef PCI_VENDOR_ID_GUILLEMOT
#define PCI_VENDOR_ID_GUILLEMOT	0x5046
#endif /* PCI_VENDOR_ID_GUILLEMOT */

#ifndef PCI_DEVICE_ID_GUILLEMOT_MAXIRADIO
#define PCI_DEVICE_ID_GUILLEMOT_MAXIRADIO	0x1001
#endif /* PCI_VENDOR_ID_GUILLEMOT_MAXIRADIO */

static void send_zero(void);
static void send_one(void);

static int stereo = 1;
static unsigned long radioport = 0;

#ifdef __FreeBSD__
static int iofd;
#endif /* __FreeBSD__ */

/*********************************************************************/

int
Grab_Port_maxiradio(unsigned long port) {
#if defined (__OpenBSD__) || defined (__NetBSD__)
	struct i386_iopl_args iopls;
#endif /* __OpenBSD__ || __NetBSD__ */

	radioport = port;

	/* FIXME: use pci code, not iopl() */
#ifdef linux /* Platform dependent */
	if ( iopl(3) < 0 )
#elif defined __FreeBSD__
	if ( (iofd = open("/dev/io", O_RDONLY)) < 0 )
#else
	iopls.iopl = 3;
	if ( sysarch(I386_IOPL, (char *)&iopls) < 0 )
#endif /* Platform dependent */
		return 1;

	return 0;
}

int
Release_Port_maxiradio(void) {
#if defined (__OpenBSD__) || defined (__NetBSD__)
	struct i386_iopl_args iopls;
#endif /* __OpenBSD__ || __NetBSD__ */

#ifdef linux /* Platform dependent */
	if ( iopl(0) < 0 )
#elif defined __FreeBSD__
	if ( close(iofd) < 0 )
#else
	iopls.iopl = 0;
	if ( sysarch(I386_IOPL, (char *)&iopls) < 0 )
#endif /* Platform dependent */
		return 1;

	return 0;
}

void
mute_maxiradio(void) {
	OUTB(radioport, 0x00);
}

void
unmute_maxiradio(void) {
	OUTB(radioport, 0x10);
}

static void
send_zero(void) {
	OUTB(radioport, 0x04);
	OUTB(radioport, 0x05);
	OUTB(radioport, 0x04);
}

static void
send_one(void) {
	OUTB(radioport, 0x06);
	OUTB(radioport, 0x07);
	OUTB(radioport, 0x06);
}

/*
 * Set frequency and other stuff
 * Basically, this is just writing the 25-bit shift register
 */
void
Change_Frequency_maxiradio(unsigned long freq) {
	int sensitivity = 5;
	int i = 15; /* Frequency bits */

	freq += 1070; /* IF 10.7 MHz */
	freq /= 1.25; /* static divisor 12.5 kHz */

	OUTB(radioport, 0x06);

	/* search start/end */
	send_zero();

	/* search up/down */
	send_zero();

	/* forced mono/stereo */
	if ( stereo )
		send_zero();
	else
		send_one();

	/* Band select - FM */
	send_zero();
	send_zero();

	/* User programmable bits */
	send_zero();
	send_zero();

	/* signal reception */
	switch ( sensitivity ) {
		case 150: /* > 150 mkV */
			send_one();
			send_one();
			break;
		case 30: /* > 30 mkV */
			send_zero();
			send_one();
			break;
		case 10: /* > 10 mkV */
			send_one();
			send_zero();
			break;
		case 5: /* > 5 mkV */
		default:
			send_zero();
			send_zero();
			break;
	}

	/* dummy */
	send_zero();

	while ( i-- )
		if ( freq & (1 << i) )
			send_one();
		else
			send_zero();

	OUTB(radioport, 0x10);
	usleep(AFC_DELAY);

	return;
}

int
state_maxiradio(void) {
	usleep(5);
	if ( inb(radioport) & 8 )
		return 3;
	return 0;
}

void
mono_maxiradio(void) {
	stereo = 0;
}
