/*
 * Copyright (c) 2000, 2001 Vladimir Popov <jumbo@narod.ru>.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
/*
 * $Id: radio.c,v 1.76 2001/07/15 12:42:12 pva Exp $
 *
 * radio.c -- implementation of interface to libradio
 *
 */

#include <stdio.h>
#include <unistd.h>
#include <string.h>

#include "radio.h"
#include "mixer.h"
#include "io.h"

static unsigned long search_up_generic(unsigned long);
static unsigned long search_down_generic(unsigned long);

#include "specs.h"

static int driver = ERADIO_INVL;
static int variant = ERADIO_INVL;

int
radio_init(char *drv) {
	int i, j;
	int errcode = ERADIO_INVL;

	int drivers = sizeof(card)/sizeof(radiocard_t);

	for ( i = 0; i < drivers; i++ ) {
		int ports = card[i].info & INFO_VARS;
		for ( j = 0; j < ports; j++ ) {
			char buf[MAX_SHNAME_LEN+1];
			int drv_len;

			if ( ports == 1 )
				snprintf(buf, MAX_SHNAME_LEN, "%s", card[i].shname);
			else
				snprintf(buf, MAX_SHNAME_LEN, "%s%d", card[i].shname, j+1);
			buf[MAX_SHNAME_LEN] = '\0';

			drv_len = strlen(buf);

			if ( strncasecmp(drv, buf, drv_len) == 0 ) {
				driver = i;
				variant =j;
				return 0;
			}
		}
	}

	return errcode;
}

unsigned long
radio_info_port(void) {
	return driver == ERADIO_INVL ? 0ul : card[driver].ports[variant];
}

char *
radio_info_name(void) {
	return driver == ERADIO_INVL ? NULL : (char *)card[driver].name;
}

int
radio_info_drv(void) {
	return driver;
}

int
radio_info_root(void) {
	return driver == ERADIO_INVL ? 0 : card[driver].info & INFO_ROOT;
}

/*
 * Scan frequency range [s_freq..e_freq], probe each frequency cycle times
 * using driver drv
 *
 * NONE -- 3
 * MONO -- 1..2
 * STEREO -- 0
 *
 * Return cummulative signal strength
 */
unsigned long
radio_scan(unsigned s_freq, unsigned e_freq, unsigned long cycle, int verbose, int drv) {
	unsigned ff, cc = 0;
	int a = 0;

	if ( drv == ERADIO_INVL )
		return 0;

	range(card[drv].lfreq, &s_freq, &e_freq, card[drv].hfreq);

	if ( e_freq == card[drv].lfreq )
		e_freq = card[drv].hfreq;

	for ( ff = s_freq; ff <= e_freq; ff++ ) {
		int i, signal = 0;

		card[drv].set_frequency(ff);

		for ( i = 0; i < cycle; i++ )
			signal += 3 - card[drv].state();

		if ( verbose )
			printf("%.2f => %d\n", (float)ff/100, signal);
		else
			draw_stick(a++);

		cc += signal;
	}

	return cc;
}

/*
 * Find ports that answer...
 */
void
radio_detect(unsigned low, unsigned high, unsigned long count) {
	int i;
	int drivers = sizeof(card)/sizeof(radiocard_t);

	puts("Probing ports, please wait...");
	for ( i = 0; i < drivers; i++ ) {
		int j;
		int ports = card[i].info & INFO_VARS;

		for ( j = 0; j < ports; j++ ) {
			int found = 0;

			if ( card[i].info & INFO_SCAN ) {
				if ( card[i].get_port(card[i].ports[j]) != 0 )
					continue; /* Uhhuh, not found */
				found = radio_scan(low, high, count, 0, i);
			} else {
				found = card[i].get_port(card[i].ports[j]) ? 0 : 1;
			}

			if ( found )  {
				if ( card[i].ports[j] == ERADIO_NONE )
					puts(card[i].name);
				else
					printf("port 0x%x: %s\n", card[i].ports[j], card[i].name);
			}
		}
		card[i].free_port();
	}
	puts("done");
}

int
radio_get_port(void) {
	return driver == ERADIO_INVL ?
		-1 : card[driver].get_port(card[driver].ports[variant]);
}

int
radio_free_port(void) {
	return driver == ERADIO_INVL ? -1 : card[driver].free_port();
}

void
radio_set_frequency(unsigned long freq) {
	if ( driver == ERADIO_INVL )
		return;

	range(card[driver].lfreq, (unsigned int *)&freq, (unsigned int *)&freq, card[driver].hfreq);

	card[driver].set_frequency(freq);
}

void
radio_mixer_mute(void) {
	mixer_set_volume("0,0");
}

void
radio_volume_up(void) {
	if ( driver == ERADIO_INVL )
		return;

	card[driver].volume_up();
}

void
radio_volume_down(void) {
	if ( driver == ERADIO_INVL )
		return;

	card[driver].volume_down();
}

void
radio_mute(void) {
	if ( driver == ERADIO_INVL )
		return;

	card[driver].mute();
}

int
radio_get_state(void) {
	if ( driver == ERADIO_INVL )
		return 3;

	return card[driver].state();
}

void
radio_set_mono(void) {
	if ( driver == ERADIO_INVL )
		return;

	/* FIXME: cards that unable to set mono should use mixer */
	card[driver].mono();
}

int
radio_mixer_set_volume(char *vol) {
	return mixer_set_volume(vol);
}

int
radio_mixer_update_volume(char *vol) {
	return mixer_update_device(vol);
}

int
radio_mixer_init() {
	return mixer_init();
}

int
radio_mixer_cleanup() {
	return mixer_cleanup();
}

unsigned long
radio_search_up(unsigned long freq) {
	if ( driver == ERADIO_INVL )
		return 0;

	range(card[driver].lfreq, (unsigned int *)&freq, (unsigned int *)&freq, card[driver].hfreq);

	freq = card[driver].search_up(freq);

	return freq;
}

unsigned long
radio_search_down(unsigned long freq) {
	if ( driver == ERADIO_INVL )
		return 0;

	range(card[driver].lfreq, (unsigned int *)&freq, (unsigned int *)&freq, card[driver].hfreq);

	freq = card[driver].search_down(freq);

	return freq;
}

static unsigned long
search_up_generic(unsigned long freq) {
	int max = 0;
	int platoe_start = 0;
	int platoe_count = 0;
	unsigned long f = freq;

	freq--;

	while ( freq < card[driver].hfreq ) {
		int c = SEARCH_PROBE;
		int s = 0;

		card[driver].set_frequency(++freq);

		while ( c-- )
			s += 3 - card[driver].state();

		/* FIXME: more precise approximation */
		if ( s > max ) {
			max = s;
			platoe_start = 1;
		} else if ( s == max ) {
			if ( platoe_start )
				platoe_count++;
		} else if ( s < max ) {
			if ( platoe_start ) {
				if ( platoe_count > SEARCH_LENGTH ) {
					freq -= 2*platoe_count/3;
					break;
				}
			} else {
				platoe_start = 0;
				platoe_count = 0;
				max = s;
			}
		}
	}

	if ( freq < card[driver].hfreq )
		card[driver].set_frequency(freq);
	else
		card[driver].set_frequency(f);

	return freq;
}

static unsigned long
search_down_generic(unsigned long freq) {
	int max = 0;
	int platoe_start = 0;
	int platoe_count = 0;
	unsigned long f = freq;

	freq++;

	while ( freq > card[driver].lfreq ) {
		int c = SEARCH_PROBE;
		int s = 0;

		card[driver].set_frequency(--freq);

		while ( c-- )
			s += 3 - card[driver].state();

		/* FIXME: more precise approximation */
		if ( s > max ) {
			max = s;
			platoe_start = 1;
		} else if ( s == max ) {
			if ( platoe_start )
				platoe_count++;
		} else if ( s < max ) {
			if ( platoe_start ) {
				if ( platoe_count > SEARCH_LENGTH ) {
					freq += platoe_count/3;
					break;
				}
			} else {
				platoe_start = 0;
				platoe_count = 0;
				max = s;
			}
		}
	}

	if ( freq > card[driver].lfreq )
		card[driver].set_frequency(freq);
	else
		card[driver].set_frequency(f);

	return freq;
}
