/*
 * Copyright (c) 2001 Vladimir Popov <jumbo@narod.ru>.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
/*
 * $Id: radiotrackII.c,v 1.24 2001/06/05 14:10:20 pva Exp $
 * lowlevel driver io functions for fmio utility -- radiotrackII driver
 */

#include <unistd.h>

#ifdef linux
#include <sys/io.h>
#elif defined __FreeBSD__
#include <fcntl.h>
#include <machine/cpufunc.h>
#else
#include <machine/pio.h>
#include <machine/sysarch.h>
#endif

#include "io.h"
#include "tea5757.h"

static struct tea5757_t card;
static unsigned long radioport = 0;
#ifdef __FreeBSD__
static int iofd;
#endif /* __FreeBSD__ */

static void send_zero(void);
static void send_one(void);
static void write_shift_register(unsigned long);
static unsigned long read_shift_register(void);

/******************************************************************/

int
Grab_Port_radiotrackII(unsigned long port) {
#if defined (__OpenBSD__) || defined (__NetBSD__)
	unsigned long iomap[32];
	int offset;
	unsigned long mask;
	struct i386_set_ioperm_args ioperms;
#endif /* __OpenBSD__ || __NetBSD__ */

	radioport = port;

#ifdef linux /* Platform dependent */
	if ( ioperm(radioport, 1, 1) < 0 )
#elif defined __FreeBSD__
	if ( (iofd = open("/dev/io", O_RDONLY)) < 0 )
#else
	offset = radioport/32;
	mask = 1 << radioport%32;
	memset(iomap, 0xFFFF, sizeof(iomap));
	iomap[offset] ^= mask;
	ioperms.iomap = iomap;
	if ( sysarch(I386_SET_IOPERM, (char *)&ioperms) < 0 )
#endif /* Platform dependent */
		return 1;

	card.frequency = 0;
	card.search = TEA5757_SEARCH_END;
	card.stereo = TEA5757_STEREO;
	card.sensitivity = TEA5757_S030;
	card.write = write_shift_register;
	card.read = read_shift_register;

	return 0;
}

int
Release_Port_radiotrackII(void) {
#if defined (__OpenBSD__) || defined (__NetBSD__)
	unsigned long iomap[32];
	struct i386_set_ioperm_args ioperms;
#endif /* __OpenBSD__ || __NetBSD__ */

#ifdef linux /* Platform dependent */
	if ( ioperm(radioport, 1, 0) < 0 )
#elif defined __FreeBSD__
	if ( close(iofd) < 0 )
#else
	memset(iomap, 0xFFFF, sizeof(iomap));
	ioperms.iomap = iomap;
	if ( sysarch(I386_SET_IOPERM, (char *)&ioperms) < 0 )
#endif /* Platform dependent */
		return 1;

	return 0;
}

void
Change_Frequency_radiotrackII(unsigned long frequency) {
	card.frequency = frequency;
	card.search = TEA5757_SEARCH_END;
	tea5757_write_shift_register(&card);
}

void
mute_radiotrackII(void) {
	OUTB(radioport, 0x01);
}

void
unmute_radiotrackII(void) {
	OUTB(radioport, 0x00);
}

int
state_radiotrackII(void) {
	int res = inb(radioport);

	if ( res == 0xfd )
		return 0;
	else
		if ( res != 0xff )
			return 2;

	return 3;
}

unsigned long
search_up_radiotrackII(unsigned long freq) {
	card.frequency = freq;
	card.search = TEA5757_SEARCH_UP;
	return tea5757_search(&card);
}

unsigned long
search_down_radiotrackII(unsigned long freq) {
	card.frequency = freq;
	card.search = TEA5757_SEARCH_DOWN;
	return tea5757_search(&card);
}

void
mono_radiotrackII(void) {
	card.stereo = TEA5757_MONO;
}

static void
send_zero(void) {
	OUTB(radioport, 0x01);
	OUTB(radioport, 0x03);
	OUTB(radioport, 0x01);
}

static void
send_one(void) {
	OUTB(radioport, 0x05);
	OUTB(radioport, 0x07);
	OUTB(radioport, 0x05);
}

static void
write_shift_register(unsigned long data) {
	int c = 25;

	OUTB(radioport, 0xc8);
	OUTB(radioport, 0xc9);
	OUTB(radioport, 0xc9);

	while ( c-- )
		if ( data & (1<<c) )
			send_one();
		else
			send_zero();

	OUTB(radioport, 0xc8);
}

static unsigned long
read_shift_register(void) {
	unsigned long reg = 0;
	int c = 25;
	int rb;
	
	OUTB(radioport, 0x06);

	while ( c-- ) {
		OUTB(radioport, 0x04);
		OUTB(radioport, 0x06);
		rb = inb(radioport);
		reg |= rb & 0x01;
		reg <<= 1;
	}

	reg >>= 1;

	return reg;
}
