/*
 * Copyright (c) 2001 Vladimir Popov <jumbo@narod.ru>.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
/*
 * $Id: sf16fmr2.c,v 1.31 2001/06/05 14:28:52 pva Exp $
 * lowlevel driver io functions for fmio utility -- sf16fmr2 driver
 */

#include <unistd.h>

#ifdef linux
#include <sys/io.h>
#elif defined __FreeBSD__
#include <fcntl.h>
#include <machine/cpufunc.h>
#else
#include <machine/pio.h>
#include <machine/sysarch.h>
#endif

#include "io.h"
#include "tea5757.h"

static struct tea5757_t card;
static unsigned long radioport = 0;
#ifdef __FreeBSD__
static int fdio;
#endif /* __FreeBSD__ */

static void send_zero(void);
static void send_one(void);
static void write_shift_register(unsigned long);
static unsigned long read_shift_register(void);

/******************************************************************/

int
Grab_Port_sf16fmr2(unsigned long port) {
#if defined (__OpenBSD__) || defined (__NetBSD__) /* __OpenBSD__  || __NetBSD__ */
	unsigned long iomap[32];
	int offset;
	unsigned long mask;
	struct i386_set_ioperm_args ioperms;
#endif /* __OpenBSD__ || __NetBSD__ */

	radioport = port;

#ifdef linux /* Platform dependent */
	if ( ioperm(radioport, 1, 1) < 0 )
#elif defined __FreeBSD__
	if ((fdio = open( "/dev/io", O_RDONLY )) < 0 )
#else
	offset = radioport/32;
	mask = 1 << radioport%32;
	memset(iomap, 0xFFFF, sizeof(iomap));
	iomap[offset] ^= mask;
	ioperms.iomap = iomap;
	if ( sysarch(I386_SET_IOPERM, (char *)&ioperms) < 0 )
#endif /* Platform dependent */
		return 1;

	/* Initialize the card */
	card.frequency = 0;
	card.search = TEA5757_SEARCH_END;
	card.sensitivity = TEA5757_S030;
	card.stereo = TEA5757_STEREO;
	card.write = write_shift_register;
	card.read = read_shift_register;

	return 0;
}

int
Release_Port_sf16fmr2(void) {
#if defined (__OpenBSD__) || defined (__NetBSD__) /* __OpenBSD__  || __NetBSD__ */
	unsigned long iomap[32];
	struct i386_set_ioperm_args ioperms;
#endif /* __OpenBSD__ || __NetBSD__ */

#ifdef linux /* Platform dependent */
	if ( ioperm(radioport, 1, 0) < 0 )
#elif defined __FreeBSD__
	if ( close(fdio) < 0 )
#else
	memset(iomap, 0xFFFF, sizeof(iomap));
	ioperms.iomap = iomap;
	if ( sysarch(I386_SET_IOPERM, (char *)&ioperms) < 0 )
#endif /* Platform dependent */
		return 1;

	return 0;
}

void
Change_Frequency_sf16fmr2(unsigned long frequency) {
	card.frequency = frequency;
	card.search = TEA5757_SEARCH_END;
	tea5757_write_shift_register(&card);
	return;
}

unsigned long
search_up_sf16fmr2(unsigned long freq) {
	card.frequency = freq;
	card.search = TEA5757_SEARCH_UP;
	return tea5757_search(&card);
}

unsigned long
search_down_sf16fmr2(unsigned long freq) {
	card.frequency = freq;
	card.search = TEA5757_SEARCH_DOWN;
	return tea5757_search(&card);
}

void
mute_sf16fmr2(void) {
	OUTB(radioport, 0x00);
}

void
unmute_sf16fmr2(void) {
	OUTB(radioport, 0x04);
}

int
state_sf16fmr2(void) {
	unsigned int res = tea5757_read_shift_register(&card);
	return (res>>25) & 0x03;
}

void
mono_sf16fmr2(void) {
	card.stereo = TEA5757_MONO;
}

static void
send_one(void) {
	OUTB(radioport, 0x01);
	OUTB(radioport, 0x03);
}

static void
send_zero(void) {
	OUTB(radioport, 0x00);
	OUTB(radioport, 0x02);
}

static unsigned long
read_shift_register(void) {
	int rb, res = 0;
	OUTB(radioport, 0x05);

	OUTB(radioport, 0x07);
	rb = inb(radioport);
	res = rb & 0x80 ? 1 : 0; /* Amp ? */
	res <<= 1;
	res |= rb & 0x08 ? 1 : 0; /* Tuned/Not tuned */
	res <<= 1;

	OUTB(radioport, 0x05);
	rb = inb(radioport);
	res |= rb & 0x08 ? 1 : 0; /* Mono/Stereo */
	res <<= 1;
	res |= rb & 0x01;

	/* Read the rest of the register */
	rb = 23;
	while ( rb-- ) {
		res <<= 1;
		OUTB(radioport, 0x07);
		OUTB(radioport, 0x05);
		res |= inb(radioport) & 0x01;
	}

	return res;
}

static void
write_shift_register(unsigned long data) {
	int c = 25;

	OUTB(radioport, 0x00);

	while ( c-- )
		if ( data & (1<<c) )
			send_one();
		else
			send_zero();

	OUTB(radioport, 0x00);
	OUTB(radioport, 0x04);

}
